"""Calculate solvent-accessible surface areas using Lee-Richards method."""
import logging
import pkg_resources
import yaml
import numpy as np
import pandas as pd
from scipy.spatial import cKDTree as Tree
from osmolytes.pqr import count_residues


_LOGGER = logging.getLogger(__name__)
# The value at which a radius is considered 0
RADIUS_CUTOFF = 0.0001
# Data files for reference states
CREAMER_DICT = yaml.load(
    pkg_resources.resource_stream(__name__, "data/creamer-areas.yaml"),
    Loader=yaml.FullLoader,
)
AUTON_DICT = yaml.load(
    pkg_resources.resource_stream(__name__, "data/denatured-areas.yaml"),
    Loader=yaml.FullLoader,
)
TRIPEPTIDE_DICT = yaml.load(
    pkg_resources.resource_stream(__name__, "data/tripeptide-areas.yaml"),
    Loader=yaml.FullLoader,
)


def sphere_cube(num):
    """Discretize a sphere using points uniformly distributed on a cube.

    Discretize a unit sphere at the origin with a cube geodesic.

    :param num:  target number of points on sphere
    :type num:  int
    :returns:  an array of (anum)-by-3 dimension where anum may not be
        exactly the same as num
    :rtype:  np.ndarray
    """
    if num < 8:
        num = 8
    num = int((np.sqrt(6 * num - 32) - 4.0) / 6.0)
    line = np.linspace(0, 1, num + 2)
    edge = np.delete(line, [0, num + 1])
    points = None
    # Generate faces
    for x in [0.0, 1.0]:
        face = np.array([(x, y_, z_) for y_ in edge for z_ in edge])
        if points is None:
            points = face
        else:
            points = np.concatenate([points, face])
    for y in [0.0, 1.0]:
        face = np.array([(x_, y, z_) for x_ in edge for z_ in edge])
        points = np.concatenate([points, face])
    for z in [0.0, 1.0]:
        face = np.array([(x_, y_, z) for x_ in edge for y_ in edge])
        points = np.concatenate([points, face])
    # Generate corners
    for x in [0.0, 1.0]:
        for y in [0.0, 1.0]:
            for z in [0, 1.0]:
                corner = np.array([x, y, z])
                points = np.vstack((points, corner))
    # Generate edges
    for x in [0.0, 1.0]:
        for y in [0.0, 1.0]:
            edge_ = np.array([(x, y, z_) for z_ in edge])
            points = np.concatenate([points, edge_])
            for z in [0, 1.0]:
                edge_ = np.array([(x, y_, z) for y_ in edge])
                points = np.concatenate([points, edge_])
                edge_ = np.array([(x_, y, z) for x_ in edge])
                points = np.concatenate([points, edge_])
    # Move center to origin
    trans = np.array([0.5, 0.5, 0.5])
    points = points - trans
    # Scale to sphere
    dist = np.linalg.norm(points, axis=1)
    points = np.divide(points, dist[:, None])
    return points


def sphere_cylinder(num):
    """Discretize a sphere using cylinder transform.

    Discretize a unit sphere at the origin.

    .. todo:: Fix duplicate points at poles

    :param num:  target number of points on sphere
    :type num:  int
    :returns:  an array of (anum)-by-3 dimension where anum may not be
        exactly the same as num
    :rtype:  np.ndarray
    """
    num = int(np.sqrt(num))
    theta = np.linspace(0, 2 * np.pi, num, endpoint=False)
    u = np.linspace(-1, 1, num)
    theta_z = np.array([(t, z) for t in theta for z in u])
    cos_theta = np.cos(theta_z[:, 0])
    sin_theta = np.sin(theta_z[:, 0])
    z2 = np.square(theta_z[:, 1])
    r = np.sqrt(1 - z2)
    x = r * cos_theta
    y = r * sin_theta
    z = theta_z[:, 1]
    return np.array([x, y, z]).T


class SolventAccessibleSurface:
    """Calculate Lee-Richards solvent-accessible surface area for folded
    proteins."""

    def __init__(
        self, atoms, probe_radius, num_points=1000, xyz_path="surface.xyz"
    ):
        """Initialize the object.

        :param list(Atom) atoms:  list of atoms from which to construct surface
        :param float probe_radius:  radius of probe atom (solvent) in Angstroms
        :param int num_points:  number of points to use for reference sphere
        :param str xyz_path:  path to xyz file (if None, no file is written)
        """
        self.atoms = atoms
        self.probe_radius = probe_radius
        self.num_points = num_points
        self.sphere = sphere_cylinder(num_points)
        self.max_radius = max([atom.radius for atom in self.atoms])
        self.max_search = 2 * self.max_radius + 2 * probe_radius
        _LOGGER.debug("max_search = %g", self.max_search)
        # Set up atom surface reference spheres
        self.surfaces = []
        for atom in self.atoms:
            if atom.radius < RADIUS_CUTOFF:
                self.surfaces.append(None)
            else:
                atom_sphere = (
                    atom.radius + self.probe_radius
                ) * self.sphere + atom.position
                self.surfaces.append(atom_sphere)
        # Set up tree structure for distance lookup
        self.tree = Tree([atom.position for atom in self.atoms])
        matrix = self.tree.sparse_distance_matrix(
            self.tree, self.max_search, output_type="coo_matrix"
        )
        # Test individual surfaces
        _LOGGER.debug(matrix)
        for i, j, distance in zip(matrix.row, matrix.col, matrix.data):
            if i != j:
                if np.isclose(distance, 0) and np.isclose(
                    self.atoms[i].radius, self.atoms[j].radius
                ):
                    errstr = f"Overlapping atoms ({i}, {j}) of equal radius"
                    _LOGGER.warning(errstr)
                    if i < j:
                        self.surfaces[i] = self.surfaces[i][0::2]
                    else:
                        self.surfaces[i] = self.surfaces[i][1::2]
                else:
                    self.prune_surface(i, j)
        # Dump surface
        if xyz_path is not None:
            self.dump_xyz(xyz_path)

    def dump_xyz(self, xyz_path):
        """Dump surface in XYZ format.

        :param str xyz_path:  path for XYZ-format data
        """
        _LOGGER.debug("Writing debug coordinates to %s", xyz_path)
        fmt = "{name} {x:>8.3f} {y:>8.3f} {z:>8.3f}"
        with open(xyz_path, "wt") as xyz_file:
            for surface in self.surfaces:
                if surface is not None:
                    for point in surface:
                        xyz_file.write(
                            "%s\n"
                            % fmt.format(
                                name="P", x=point[0], y=point[1], z=point[2]
                            )
                        )

    def atom_surface_area(self, iatom):
        """Calculate surface area for this atom.

        :param int iatom:  index of the atom in the atom list
        :returns:  total surface area (Angstroms^2)
        :rtype:  float
        """
        num_ref = np.shape(self.sphere)[0]
        surf = self.surfaces[iatom]
        if surf is not None:
            num_surf = np.shape(surf)[0]
        else:
            num_surf = 0
        atom = self.atoms[iatom]
        tot_radius = atom.radius + self.probe_radius
        area = 4 * np.pi * tot_radius * tot_radius
        return area * float(num_surf) / float(num_ref)

    def surface_area_dictionary(self):
        """Calculate surface area, indexed by chain/residue.

        :returns:  surface area (Angstroms^2) for each residue
        :rtype:  dict
        """
        area_dict = {}
        for iatom, atom in enumerate(self.atoms):
            if atom.chain_id:
                chain_id = f"{atom.chain_id}:"
            else:
                chain_id = ""
            key = f"{chain_id}{atom.res_num}:{atom.res_name}"
            area = self.atom_surface_area(iatom)
            if key in area_dict:
                area_dict[key] = area_dict[key] + area
            else:
                area_dict[key] = area
        return area_dict

    def prune_surface(self, iatom1, iatom2):
        """Prune the surface of atom1 based on the presence of atom2.

        :param int iatom1:  index of first atom
        :param int iatom2:  index of second atom
        """
        atom2 = self.atoms[iatom2]
        if self.surfaces[iatom1] is not None:
            disp12 = self.surfaces[iatom1] - atom2.position
            dist12 = np.sum(disp12 ** 2, axis=1)
            max12 = np.square(atom2.radius + self.probe_radius)
            self.surfaces[iatom1] = self.surfaces[iatom1][dist12 > max12]


class ReferenceModels:
    """Calculate solvent-accessible surface area for reference states.

    Uses data and models from:

    - Creamer TP, Srinivasan R, Rose GD. Modeling unfolded states of proteins
      and peptides. II. Backbone solvent accessibility. Biochemistry. 1997 Mar
      11; 36(10):2832-5. doi: 10.1021/bi962819o. PMID: 9062111.
    - Auton M, Bolen DW. Predicting the energetics of osmolyte-induced protein
      folding/unfolding. Proc Natl Acad Sci U S A. 2005 Oct 18;102(42):
      15065-8. doi: 10.1073/pnas.0507053102. Epub 2005 Oct 7. PMID: 16214887;
      PMCID: PMC1257718.
    """

    def __init__(self):
        self._creamer_dict = CREAMER_DICT
        self._auton_dict = AUTON_DICT
        self._tripeptide_dict = TRIPEPTIDE_DICT

    def residue_area(self, residue, model, how="mean"):
        """Return reference area for the given residue type.

        :param str residue:  3-letter residue name
        :param str model:  model to use

          - creamer: denatured areas from Creamer, Srinivasan, Rose.
            doi: 10.1021/bi962819o
          - auton: denatured areas from Auton, Bolen
            doi: 10.1073/pnas.0507053102
          - tripeptide:  Gly-X-Gly areas for residue X from Auton, Bolen
            doi: 10.1073/pnas.0507053102

        :param str how:  calculation to perform on model values

          - max
          - min
          - mean

        :returns:  model values for sidechain and backbone
        :rtype:  dict
        """
        residue = residue.upper()
        model = model.lower()
        how = how.lower()
        areas = {}
        for what in ["sidechain", "backbone"]:
            if model == "creamer":
                values = self._creamer_dict[residue][what]
            elif model == "auton":
                values = [self._auton_dict[residue][what]]
            elif model == "tripeptide":
                values = [self._tripeptide_dict[residue][what]]
            else:
                err = f"Unknown model: {model}"
                raise ValueError(err)
            if how == "max":
                value = np.max(values)
            elif how == "min":
                value = np.min(values)
            elif how == "mean":
                value = np.mean(values)
            else:
                err = f"Unknown calculation: {how}"
                raise ValueError(err)
            areas[what] = value
        return areas

    def molecule_areas(self, atoms, model, how="mean"):
        """Return reference areas for the molecule.

        :param list(pqr.Atom) residue:  list of atoms in molecule
        :param str model:  model to use

          - creamer: denatured areas from Creamer, Srinivasan, Rose.
            doi: 10.1021/bi962819o
          - auton: denatured areas from Auton, Bolen
            doi: 10.1073/pnas.0507053102
          - tripeptide:  Gly-X-Gly areas for residue X from Auton, Bolen
            doi: 10.1073/pnas.0507053102

        :param str what:  what area to report

          - sidechain
          - backbone
          - total

        :param str how:  calculation to perform on model values

          - max
          - min
          - mean

        :returns:  DataFrame with a row for each residue type
        :rtype:  pd.DataFrame
        """
        count = count_residues(atoms)
        rows = {}
        for res, num in count.iteritems():
            res_areas = self.residue_area(res, model, how)
            row = {}
            for what in ["sidechain", "backbone"]:
                row[what] = num * res_areas[what]
            rows[res] = row
        return pd.DataFrame(rows).T.sort_index()



"""PQR file class.

Parse and store parametrized molecular structure files.
"""
import logging
from collections import Counter, OrderedDict
import numpy as np
import pandas as pd


_LOGGER = logging.getLogger(__name__)


BACKBONE_ATOMS = {"OXT", "H2", "H3", "H", "C", "O", "N", "HA", "HA2", "CA"}
SIDECHAIN_ATOMS = {
    "CB",
    "HH21",
    "OE1",
    "HZ",
    "HD22",
    "HZ1",
    "HH22",
    "CZ",
    "OE2",
    "NE",
    "CE1",
    "CD",
    "HG2",
    "HB1",
    "HD12",
    "NE2",
    "HG12",
    "SG",
    "HH11",
    "HB",
    "HE21",
    "HD23",
    "CG1",
    "HA3",
    "HE22",
    "HD11",
    "OH",
    "HG23",
    "HE",
    "HH",
    "HD21",
    "SD",
    "OG",
    "HZ3",
    "HB2",
    "HD3",
    "NH1",
    "HH12",
    "HD2",
    "HG1",
    "NZ",
    "HE2",
    "OD1",
    "HB3",
    "ND1",
    "CG2",
    "HG3",
    "HE3",
    "HG13",
    "CE",
    "OD2",
    "HE1",
    "HG21",
    "NH2",
    "HD13",
    "CD2",
    "HG22",
    "OG1",
    "CD1",
    "HG",
    "CE2",
    "ND2",
    "HZ2",
    "HD1",
    "CG",
    "HG11",
    "CZ2",
    "CE3",
    "CZ3",
    "NE1",
    "HH2",
    "CH2",
}


class Atom:
    """Atom class for storing PQR information."""

    def __init__(self, pqr_string=None):
        """Initialize Atom class.

        :param str pqr_string:  line from PQR file for initializing structure
        """
        self.pqr_string = pqr_string
        self.entry_type = None
        self.pqr_atom_num = None
        self.atom_name = None
        self.res_name = None
        self.chain_id = None
        self.res_num = None
        self.ins_code = None
        self.position = None
        self.charge = None
        self.radius = None
        self.neighbors = []
        if pqr_string is not None:
            self.parse_pqr(pqr_string)

    def parse_pqr(self, pqr_string):
        """Parse PQR string for atom.

        .. todo:: look for charge variations not captured by naming scheme
                  (e.g., ASH vs. ASP)

        :param str pqr_string:  line from PQR file for initializing structure
        """
        line = pqr_string.strip()
        words = line.split()
        if len(words) > 11:
            errstr = "Too many entries ({num}) in PQR file line: {line}"
            raise ValueError(errstr.format(num=len(words), line=line))
        elif len(words) < 10:
            errstr = "Too few entries ({num}) in PQR file line: {line}"
            raise ValueError(errstr.format(num=len(words), line=line))
        elif len(words) == 11:
            has_chain = True
        else:
            has_chain = False
        self.entry_type = words[0]
        self.pqr_atom_num = int(words[1])
        self.atom_name = words[2]
        self.res_name = words[3]
        if has_chain:
            self.chain_id = words[4]
            chain_offset = 1
        else:
            chain_offset = 0
        self.res_num = int(words[4 + chain_offset])
        x = float(words[5 + chain_offset])
        y = float(words[6 + chain_offset])
        z = float(words[7 + chain_offset])
        self.position = np.array([x, y, z])
        self.charge = float(words[8 + chain_offset])
        self.radius = float(words[9 + chain_offset])

    def __str__(self):
        outstr = ""
        tstr = self.entry_type
        outstr += str.ljust(tstr, 6)[:6]
        tstr = f"{self.pqr_atom_num:d}"
        outstr += str.rjust(tstr, 5)[:5]
        outstr += " "
        tstr = self.atom_name
        if len(tstr) == 4 or len(tstr.strip("FLIP")) == 4:
            outstr += str.ljust(tstr, 4)[:4]
        else:
            outstr += " " + str.ljust(tstr, 3)[:3]
        tstr = self.res_name
        if len(tstr) == 4:
            outstr += str.ljust(tstr, 4)[:4]
        else:
            outstr += " " + str.ljust(tstr, 3)[:3]
        outstr += " "
        tstr = self.chain_id if self.chain_id else ""
        outstr += str.ljust(tstr, 1)[:1]
        tstr = f"{self.res_num:d}"
        outstr += str.rjust(tstr, 4)[:4]
        outstr += f"{self.ins_code}   " if self.ins_code else "    "
        tstr = f"{self.position[0]:8.3f}"
        outstr += str.ljust(tstr, 8)[:8]
        tstr = f"{self.position[1]:8.3f}"
        outstr += str.ljust(tstr, 8)[:8]
        tstr = f"{self.position[2]:8.3f}"
        outstr += str.ljust(tstr, 8)[:8]
        ffcharge = (
            f"{self.charge:.4f}" if self.charge is not None else "0.0000"
        )
        outstr += str.rjust(ffcharge, 8)[:8]
        ffradius = (
            f"{self.radius:.4f}" if self.radius is not None else "0.0000"
        )
        outstr += str.rjust(ffradius, 7)[:7]
        return outstr

    def distance2(self, other):
        """Return the squared distance between this atom and another

        :param Atom other:  other atom for computing distance
        :returns:  squared distance
        :rtype:  float
        """
        displacement = self.position - other.position
        return np.inner(displacement, displacement)


def count_residues(atoms):
    """Count residues by type

    :param list(Atom) atoms:  array of atoms to count
    :returns:  Series with counts
    :rtype:  pd.Series
    """
    res_set = {(atom.res_num, atom.res_name) for atom in atoms}
    counts = Counter()
    for res_tup in res_set:
        res_type = res_tup[1]
        counts.update([res_type])
    return pd.Series(counts).sort_index()


def aggregate(
    atoms, data, chain_id=True, res_name=True, res_num=True, sidechain=True
):
    """Aggregate the array of data into a dictionary.

    .. todo::  This is a horrible kludge.

    :param list(Atom) atoms:  array of atoms over which to aggregate data
    :param list(float) data:  array of data to aggregate
    :param bool chain_id:  whether to aggregate by chain
    :param bool res_name:  whether to aggregate by residue type
    :param bool res_num:  whether to aggregate by residue number
    :param bool sidechain:  whether to aggregate by sidechain/backbone
    :returns:  DataFrame with aggregated data
    :rtype:  pd.DataFrame
    """
    aggregate = OrderedDict()
    unknown_atoms = set()
    unknown_error = False
    for iatom, atom in enumerate(atoms):
        key = []
        if chain_id:
            key.append(atom.chain_id)
        if res_name:
            key.append(atom.res_name)
        if res_num:
            key.append(str(atom.res_num))
        if sidechain:
            if atom.atom_name in BACKBONE_ATOMS:
                key.append("backbone")
            elif atom.atom_name in SIDECHAIN_ATOMS:
                key.append("sidechain")
        key = " ".join(key)
        if key in aggregate:
            aggregate[key] = aggregate[key] + data[iatom]
        else:
            aggregate[key] = data[iatom]
    if unknown_error:
        err = f"Unknown atom types:  {unknown_atoms}"
        raise ValueError(err)
    rows = []
    for key, value in aggregate.items():
        row = key.split()
        row.append(value)
        rows.append(row)
    columns = []
    if chain_id:
        columns.append("chain_id")
    if res_name:
        columns.append("res_name")
    if res_num:
        columns.append("res_num")
    if sidechain:
        columns.append("sidechain")
    columns.append("value")
    df = pd.DataFrame(rows, columns=columns)
    return df


def parse_pqr_file(pqr_file):
    """Parse a PQR file.

    :param file pqr_file:  input file-like object ready for reading
    :returns:  list of atom objects
    :rtype:  list(Atom)
    """
    atoms = []
    for line in pqr_file:
        line = line.strip()
        if line:
            words = line.split()
            if words[0] in [
                "HEADER",
                "REMARK",
                "TER",
                "COMPND",
                "AUTHOR",
                "END",
            ]:
                pass
            elif words[0] in ["ATOM", "HETATM"]:
                atoms.append(Atom(line))
            else:
                errstr = f"Unable to parse PQR line: {line}"
                raise ValueError(errstr)
    return atoms



"""Osmolytes and protein stability

This code attempts to predict the influence of osmolytes on protein stability,
using the methods in the following paper:

* Auton M, Bolen DW. Predicting the energetics of osmolyte-induced protein
  folding/unfolding. *Proc Natl Acad Sci* 102:15065 (2005)
  https://doi.org/10.1073/pnas.0507053102
"""



"""Set up the osmolytes package."""
import setuptools


with open("README.md", "r") as readme:
    LONG_DESCRIPTION = readme.read()


setuptools.setup(
    name="osmolytes",
    version="0.0.1",
    description=(
        "This code attempts to predict the influence of osmolytes on protein "
        "stability"
    ),
    long_description=(
        "This code attempts to predict the influence of osmolytes on protein "
        "stability, using the methods in:  Auton M, Bolen DW. Predicting the "
        "energetics of osmolyte-induced protein folding/unfolding. "
        "*Proc Natl Acad Sci* 102:15065 (2005) "
        "https://doi.org/10.1073/pnas.0507053102. Other models may be added "
        "in the future."
    ),
    python_requires=">=3.6",
    license="CC0-1.0",
    author="Nathan Baker",
    author_email="nathanandrewbaker@gmail.com",
    url="https://github.com/Electrostatics/osmolytes",
    packages=setuptools.find_packages(),
    package_data={
        "": ["data/*.yaml", "tests/data/*.json", "tests/data/*.yaml"]
    },
    install_requires=["numpy", "scipy", "pyyaml", "pandas"],
    tests_require=["pytest"],
    entry_points={"console_scripts": ["mvalue=osmolytes.main:main",]},
    keywords="science chemistry biophysics biochemistry",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Common Public License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)



"""Transfer energy calculation from Auton & Bolen.

From Eqs. 1 and 2 in Supporting Text from:

  Auton M, Bolen DW. Predicting the energetics of osmolyte-induced
  protein folding/unfolding. Proc Natl Acad Sci U S A. 2005 Oct 18;102(42):
  15065-8. doi: 10.1073/pnas.0507053102. Epub 2005 Oct 7. PMID: 16214887;
  PMCID: PMC1257718.
"""
import logging
import pkg_resources
import yaml
import pandas as pd
from osmolytes.pqr import aggregate, count_residues
from osmolytes.sasa import ReferenceModels


_LOGGER = logging.getLogger(__name__)
ENERGY_DICT = yaml.load(
    pkg_resources.resource_stream(__name__, "data/transfer-energies.yaml"),
    Loader=yaml.FullLoader,
)


def get_folded_areas(atoms, sas):
    """Get a summary by residue type of folded areas.

    .. todo:: Move to the :mod:`sasa` module.

    :param list(pqr.Atom) atoms:  list of atoms
    :param SolventAccessibleSurface sas: solvent-accessible surface area object
    :returns:  DataFrame with folded sidechain/backbone area by residue type
    :rtype:  pd.DataFrame
    """
    folded_areas = [
        sas.atom_surface_area(iatom) for iatom in range(len(atoms))
    ]
    folded_areas = aggregate(
        atoms,
        folded_areas,
        chain_id=False,
        res_name=True,
        res_num=False,
        sidechain=True,
    ).set_index("res_name")
    folded_areas = folded_areas.rename({"sidechain": "what"}, axis="columns")
    grouper = folded_areas.groupby("what")
    folded_areas = pd.concat(
        [pd.Series(v["value"], name=k) for k, v in grouper], axis="columns"
    ).sort_index()
    return folded_areas


def transfer_energy(atoms, sas):
    """Calculate transfer energy for each osmolyte and amino acid type.

    .. todo:: implement max/min for Creamer model

    :param list(pqr.Atom) atoms:  list of atoms
    :param SolventAccessibleSurface sas: solvent-accessible surface area object
    :returns:  DataFrame with energy contribution (kcal/mol) by residue type
    :rtype:  pd.DataFrame
    """
    ref_model = ReferenceModels()
    counts = count_residues(atoms)
    folded_areas = get_folded_areas(atoms, sas)
    unfolded_areas = ref_model.molecule_areas(
        atoms, model="auton", how="mean"
    )
    tripeptide_areas = ref_model.molecule_areas(
        atoms, model="tripeptide", how="mean"
    )
    scaled_areas = (unfolded_areas - folded_areas) / tripeptide_areas
    scaled_areas["sidechain"]["GLY"] = 0.0
    osmolytes = list(ENERGY_DICT["backbone"].keys())
    residue_energies = {}
    bb_energy = {osmolyte: 0.0 for osmolyte in osmolytes}
    for res, area in scaled_areas.T.iteritems():
        sc_energy = {}
        for osmolyte in osmolytes:
            sc_energy[osmolyte] = (
                counts[res] * ENERGY_DICT[res][osmolyte] * area["sidechain"]
            )
            bb_energy[osmolyte] += (
                counts[res]
                * ENERGY_DICT["backbone"][osmolyte]
                * area["backbone"]
            )
        residue_energies[res] = sc_energy
    residue_energies["backbone"] = bb_energy
    return pd.DataFrame(residue_energies).T.sort_index() / 1000



"""Main driver for module."""
import argparse
import logging
from sys import version_info
from osmolytes.pqr import parse_pqr_file, count_residues
from osmolytes.sasa import SolventAccessibleSurface
from osmolytes.energy import transfer_energy


assert version_info > (3, 5)
_LOGGER = logging.getLogger("main")


def build_parser():
    """Build argument parser.

    :returns:  argument parser
    :rtype:  argparse.ArgumentParser
    """
    parser = argparse.ArgumentParser(
        description=(
            "This code uses various models to attempt to predict the "
            "influence of osmolytes on protein stability"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--log-level",
        help="set verbosity of code",
        nargs=1,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "--model",
        help=(
            "model to use for stability prediction:\n"
            "  'auton' -- Auton M, Bolen DW. Predicting the energetics of "
            "osmolyte-induced protein folding/unfolding. Proc Natl Acad Sci "
            "102:15065 (2005) https://doi.org/10.1073/pnas.0507053102"
        ),
        nargs=1,
        default="auton",
        choices=["auton"],
    )
    parser.add_argument(
        "--surface-output",
        nargs=1,
        default=None,
        help="path for output of surface in XYZ format",
    )
    parser.add_argument(
        "--solvent-radius",
        nargs=1,
        default=1.4,
        type=float,
        help="radius of solvent molecule (in Å)",
    )
    parser.add_argument(
        "--surface-points",
        nargs=1,
        default=2000,
        type=int,
        help="number of points per atom for constructing surface",
    )
    parser.add_argument(
        "pqr_path",
        help=(
            "input PQR file for protein, as generated by PDB2PQR or "
            "similar program (see documentation) "
        ),
    )
    return parser


def main():
    """Main driver."""
    parser = build_parser()
    args = parser.parse_args()
    logging.basicConfig(level=getattr(logging, args.log_level))
    _LOGGER.debug(f"Got command-line arguments: {args}")
    _LOGGER.info(f"Reading PQR input from {args.pqr_path}...")
    with open(args.pqr_path, "rt") as pqr_file:
        atoms = parse_pqr_file(pqr_file)
    count_df = count_residues(atoms)
    _LOGGER.info(f"Protein composition:\n{count_df}")
    _LOGGER.info(
        f"Constructing protein solvent-accessible surface with "
        f"solvent radius {args.solvent_radius} and {args.surface_points} "
        f"points per atom."
    )
    sas = SolventAccessibleSurface(
        atoms,
        probe_radius=args.solvent_radius,
        num_points=args.surface_points,
        xyz_path=args.surface_output,
    )
    _LOGGER.info("Calculating transfer energies.")
    energy_df = transfer_energy(atoms, sas)
    _LOGGER.info(f"Detailed energies (kcal/mol/M):\n{energy_df.to_string()}")
    energies = energy_df.sum(axis=0).sort_values()
    _LOGGER.info(f"Summary energies (kcal/mol/M):\n{energies}")
